#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#ifdef __APPLE__
#include <strings.h>
#else
#include <memory.h>
#endif
#ifndef DEBUG
#include <qdos.h>
#else
#include <sys/types.h>
#endif

#include "../include/m4v.h"

#define DISKBUFSIZ 32*1024

#ifdef DEBUG
#define TESTFILE "test_m4v"
#else
#define TESTFILE "ram1_test_m4v"
#endif

#ifdef DEBUG
unsigned char screen[256][128];
#else
unsigned char *screen = (unsigned char *) 0x20000;
#endif

int buffer_read(int fd, void *buf, size_t size);

int main()
{
	int fd;
	int i, j, k, count;
	int cur_line = 0;
	unsigned char databyte;
	unsigned char *address;
	struct difference {
		unsigned char address;
		unsigned char value;
	} diffdat;
	unsigned int value;
	unsigned char line_buffer[128];

#ifndef DEBUG
	unsigned char *scr_line_ptrs[256];

	for (i = 0; i < 256; i++)
	{
		scr_line_ptrs[i] = (unsigned char *)(screen + (128 * i));
		for (j = 0; j < 128; j++)
		{
			*(scr_line_ptrs[i] + j) = (unsigned char) 0x00;
		}
	}
#else
	unsigned char *scr_line_ptrs[256];

	for (i = 0; i < 256; i++)
		scr_line_ptrs[i] = screen[i];
		
#endif

	if ((fd = open(TESTFILE, O_RDONLY)) < 0)
	{
		fprintf(stderr, "Can't open %s\n", TESTFILE);
		return -1;
	}

	while((buffer_read(fd, &databyte, sizeof(unsigned char))) == sizeof(unsigned char))
	{
#ifdef DEBUG
		fprintf(stderr, "databyte is 0x%02x\n", databyte);
#endif
		if (isnewframe(databyte) == 1)
		{
			cur_line = 0;
#ifndef DEBUG
			mt_susjb(-1,1,NULL);
#else
			fprintf(stderr, "***************************** NEW BALLS PLEASE! ********************************\n");
#endif
		}
		else
		{
			if (cur_line == 256)
			{
				fprintf(stderr, "ERROR! got a frame with more than 256 lines!\n");
				return -1;
			}
		}

#ifdef DEBUG
		fprintf(stderr, "Line number: %d: \n", cur_line);
#endif

		if (isdiffline(databyte) == 1)
		{
			count = (databyte & 0x3f);
#ifdef DEBUG
			fprintf(stderr, "We have a differential line with %d changes.\n", count);
#endif
			for (i = 0; i < count; i++)
			{
				if (buffer_read(fd, &diffdat, sizeof(struct difference)) != sizeof(struct difference))
				{
					fprintf(stderr, "Bad read of difference block.\n");
					return -1;
				}
#ifdef DEBUG
				fprintf(stderr, "Change at line %d, block %d: %02x\n", cur_line, diffdat.address, diffdat.value);
#endif
				address = (unsigned char *)(scr_line_ptrs[cur_line] + (diffdat.address * 2));
				*address = diffdat.value;
				*(address + 1) = diffdat.value;
			}
		}
		else
		{
#ifdef DEBUG
			fprintf(stderr, "Got a full scanline.\n");
#endif
			for (i = 0; i < 128; i++)
			{
				if (buffer_read(fd, &databyte, sizeof(unsigned char)) != sizeof(unsigned char))
				{
					fprintf(stderr, "Misread of fill line data.\n");
					return -1;
				}
				line_buffer[i++] = (unsigned char)((unsigned char) databyte);
				line_buffer[i] = (unsigned char)((unsigned char) databyte);
			}

			bcopy((char *)line_buffer, (char *)scr_line_ptrs[cur_line], 128);
		}

		cur_line++;
		databyte = 0x00;
	}
#ifdef DEBUG
	fprintf(stderr, "End of file, I think.\n");
#endif
	close(fd);
}

static char buffer[DISKBUFSIZ];
static unsigned long bufpos = 0, bufend = 0;

int buffer_read(int fd, void *dest, size_t reqsiz)
{
	unsigned long readbytes, bytesleft, numbytes;

/* Check to see if we have enough data in the buffer to fulfil
	the request. If not then move the remainder of the data
	to the beginning of the buffer.
*/

	if ((bufpos + reqsiz) >= bufend)
	{
#ifdef DEBUG
		fprintf(stderr, "buffer_read: not enough data in buffer: bufpos(%lu) reqsiz(%lu) bufend(%lu)\n",
			bufpos, reqsiz, bufend);
#endif
			
		bytesleft = bufend - bufpos;

		if (bytesleft > 0)
		{
#ifdef DEBUG
			fprintf(stderr, "buffer_read: There are still %lu bytes left in the buffer.\n", bytesleft);
#endif
			memcpy(buffer, &buffer[bufpos], bytesleft);
		}

		bufpos = 0;
		bufend = bytesleft;

		readbytes = read(fd, &buffer[bufend], (DISKBUFSIZ - 1) - bufend);

		bufend = bufend + readbytes;
#ifdef DEBUG
		fprintf(stderr, "buffer_read: Read %lu bytes from the file, there are now %lu bytes in the buffer.\n", readbytes, bufend);
#endif
	}

	if ((bufend - bufpos) > reqsiz)
		numbytes = reqsiz;
	else
		numbytes = (bufend - bufpos);
		
	bcopy(&buffer[bufpos], dest, numbytes);

	bufpos = bufpos + numbytes;

#ifdef DEBUG
	fprintf(stderr, "buffer_read: Copied %lu bytes into the provided memory, new buffer position is %lu\n", numbytes, bufpos);
#endif

	return numbytes;
}
